//! A library with reusable helpers and types for the `wkg` CLI. This is intended to be used by
//! other downstream CLI tools that may need to leverage some of the same functionality provided by
//! `wkg`.

pub mod lock;
pub mod manifest;
pub mod resolver;
pub mod wit;
